//CPAGE================================================================================
//
// Demo for Nortel / Noritake 20x2 VFD Display
// =========================================== 
// 
// Copyright:
// ----------
//  Public Domain
//
// Version History:
// ----------------
//  V1.0 - 2014-12-05 - Initial code.
//                    - Development Environment: Arduino V1.06
//
// Overview:
// ---------
// Run a Noritake VFD module from a Nortel Millenium pay phone through many of its paces.
// - A base to develop a fully functional driver.
// - Testing environment to discover features of the module.
//
// Hardware:
// ---------
//   CPU board: Solarbotics Freeduino V2.1 (old version)
//   - Atmel ATmega128 @ 16MHz, 16K flash, 1K RAM, 512 bytes EEPROM.
//   Configure as: "Arduino Diecimala or Dumilanove W/ATmega168" 
//   Serial monitor: 115200
//   
//===================================================================================
#include <stdlib.h>

//CPAGE========================================================================================
// Pin definitions
//
// - The following pins must be connected to the VFD as follows in the hardware setup.
// - Any Arduino digital capabile pin may be used for any function.
// - All port manipulation is performed by high-level Arduino functions.
//=============================================================================================
//       Software  Digital    Hardware             Display  Display 
//         Name     Pin#       Pin-ID      Mode     Pin#    Pin-Description
#define   VFD_D7     2     //    D2        Out       15       D0    - Data bus
#define   VFD_D6     3     //    D3        Out       13       D1
#define   VFD_D5     4     //    D4        Out       11       D2
#define   VFD_D4     5     //    D5        Out        9       D3
#define   VFD_D3     6     //    D6        Out        7       D4
#define   VFD_D2     7     //    D7        Out        5       D5
#define   VFD_D1     8     //    D8        Out        3       D6
#define   VFD_D0     9     //    D9        Out        1       D7    - Data bus
#define   VFD__WR   14     //    D14-A0    Out       17       _WR   - Write
#define   VFD_AD    15     //    D15-A1    Out       19       AD    - Address/Data mode
#define   VFD__RD   16     //    D16-A2    Out       21       ST    - Read mode
#define   VFD__CS   17     //    D17-A3    Out       23       _CS   - Chip select
#define   VFD__TS   18     //    D18-A4    Out       25       _TS   - Test mode  
#define   VFD_RS    19     //    D19-A5    Out       20       RS    - Hardware reset

//CPAGE========================================================================================
// VFD Control Constants
//=============================================================================================
// Display parameters and cursor positioning constants used in command mode
#define VFD_NumLines                 2    // 2 lines
#define VFD_LineLength              20    // 20 characters per line
#define VFD_Addr_L1_First            0    // Cursor position - Top-left
#define VFD_Addr_L2_First           20    // Cursor position - Bot-Left
#define VFD_Addr_L1_Last            19    // Cursor position - Top-Right
#define VFD_Addr_L2_Last            39    // Cursor position - Bot-Right 

#define VFD_Cmd_SWReset           0x40    // Software reset - from cmd mode only
#define VFD_Cmd_Read_Pos          0x41    // Read cursor posn - this may not work

// Control codes that can be sent to the display in data mode
// - See documentation for more details.
#define VFD_Ctl_LF                0x0A    // Move cursor down one line      
#define VFD_Ctl_FF                0x0C    // Clear display, does not move cursor

#define VFD_Mode_Normal           0x11    // DC1 Normal display mode
#define VFD_Mode_VertScroll       0x12    // DC2 Sroll vertically 
#define VFD_Mode_Cursor_On        0x13    // DC3 Turn cursor on
#define VFD_Mode_Cursor_Off       0x14    // DC4 Turn cursor off
#define VFD_Mode_HorzScroll       0x15    // DC5 Scroll horizontially
#define VFD_Mode_Blinking_On      0x16    // DC6 Blink a corner positions (2-byte code)
#define VFD_Mode_Blinking_Off     0x17    // DC7 Turn off all blinking

#define VFD_Set_Brightness        0x1A    // set intensity level (2-byte code)
#define VFD_Set_Char_Bitmap       0x1B    // ESC - Define bitmap (7-byte code)

// Time in microseconds for the display to process variopus types of command.
// - These are very conservative top allow for worst-case situations. Thay come from 
//   analysis of the display being used in an application.
// - Do you own testing to determine smaller values for your application
//   in specific write sequences if speed matters that much to you.
// - Many functions worked fine down to 10-20us delays between commands.
// - It would be better to monitor a busy flag, but it does not seem that the
//   display provides that status information.
// - The display is much faster than HD44780 type LCD displays that need to support a busy flag.
// - This sample code uses a simplistic set of delays.
//   - Cmds are either a SWReset or a Cursor_Posn
//   - Data is either VFD_Ctl_FF (clear) or something else for which VFD_Delay_Mode_Normal will do.
#define VFD_Hold_Test           150000    // Hold test line low for 150ms to initiate.
#define VFD_Delay_Test          100000    // Allow 100ms for test to initiate.
#define VFD_Delay_HWReset       150000    // Allow 150ms for HW resets to complete.
#define VFD_Delay_SWReset       100000    // Allow 100ms for SW resets to complete.
#define VFD_Delay_Reset_Pulse        5    // Hold reset line for this long to reset.
#define VFD_Delay_Cursor_Posn       40    // Move the cursor position
#define VFD_Delay_Clear            560    // Clear the display
#define VFD_Delay_DC1              100    // Change mode to normal mode
#define VFD_Delay_Mode_Normal      120    // Write a character in normal mode
#define VFD_Delay_Cursor_Mode      100    // Change cursor mode: DC3, DC4, DC5, DC6, DC7
#define VFD_Delay_Set_Blink_Posn    70    // Set a position to blinking mode
#define VFD_Delay_Mode_HorizScroll  90    // Write a character in horizontal scrolling mode

// Constants used to set the brightness of the display.
#define VFD_Brightness_100        0x00    // 100% 
#define VFD_Brightness_75         0x10    //  75% 
#define VFD_Brightness_50         0x20    //  50%
#define VFD_Brightness_25         0x30    //  25%

//CPAGE========================================================================================
// Global Data
//=============================================================================================
unsigned long     LastCmdTime=0;     // micros() timer value when last command was executed.
unsigned long     VFDCmdWait=500;    // us to wait until the next command may be issued.
char              VFD_Buffer[40];    // Full display buffer for VFD. 

//CPAGE========================================================================================
// setup
// - Executed once after device reset.
//=============================================================================================
void setup() {
  Serial.begin(115200);               // Make sure PC virtual serial port is set to this.
  Serial.setTimeout(1);               // Don't wait for stuff!
  Serial.println("Starting . . . .");

  // Set up pins for VFD communications
  pinMode(VFD_D0,  OUTPUT);
  pinMode(VFD_D1,  OUTPUT);
  pinMode(VFD_D2,  OUTPUT);
  pinMode(VFD_D3,  OUTPUT);
  pinMode(VFD_D4,  OUTPUT);
  pinMode(VFD_D5,  OUTPUT);
  pinMode(VFD_D6,  OUTPUT);
  pinMode(VFD_D7,  OUTPUT);
  pinMode(VFD__WR, OUTPUT);
  pinMode(VFD_AD,  OUTPUT);
  pinMode(VFD__RD, OUTPUT);  
  pinMode(VFD__CS, OUTPUT);
  pinMode(VFD__TS, OUTPUT);
  pinMode(VFD_RS,  OUTPUT); 
  
  // Start the initial states of the control lines.
  // - The data lines don't matter at this point.
  digitalWrite(VFD__WR, HIGH);   // _WR - Write mode - Active low)
  digitalWrite(VFD_AD,  LOW);    // AD  - Address/Data mode - Always data?
  digitalWrite(VFD__RD, HIGH);   // RD  - Read mode: 0-Data/Status Read 1-Data/Cmd Write
  digitalWrite(VFD__CS, LOW);    // _CS - Chip select - Always selected?
  digitalWrite(VFD__TS, HIGH);   // _TS - Test mode bit - active low
  digitalWrite(VFD_RS,  LOW);    // RS  - hardware reset line

  LastCmdTime=0;                // We have not issued a command yet, so no need to wait.
  VFD_HW_Reset();               // Do a hardware reset just because we can!
  
} //  Setup
    
//CPAGE========================================================================================
// loop
// - executed in an endless loop until the device is reset or powered down.
//=============================================================================================
void loop() {
  
  //CPAGE--------------------------------------------------------------------------------------
  //  Test mode and resets
  //-------------------------------------------------------------------------------------------
  
  // Put the display into test mode for while. The only way to cancel it is with a device hardware reset.
  // - A software reset will not reset test mode.
  VFD_Test_Mode();                               // Go into test mode
  delay(5000);                                   // Wait to see some of the display

  // Now do a hardware reset to clear test mode. It is the only way out of test.
  VFD_HW_Reset();            

  //CPAGE--------------------------------------------------------------------------------------
  //  Display updates via full screen buffer
  // - With the cuirsor on you can see it moving as the display is updated. 
  // - The display looks much better with cursor off. 
  // - It takes about 7ms to do all of the work in the loop, so although the counter decrements
  //   by 25, we only wait 18ms at the end.     
  //-------------------------------------------------------------------------------------------
  char Num_Buff[6];
  
  // User full screen updates to dislay a fast moving count down.
  memset(VFD_Buffer, ' ', sizeof(VFD_Buffer));
  VFD_Str_To_Buffer(VFD_Buffer,    "Buffer demo:");
  VFD_Str_To_Buffer(VFD_Buffer+20 ,"xxxxx ms to go!");
  VFD_Send_Char(VFD_Mode_Cursor_Off);
  VFD_Send_Cmd(20);
  for (int i = 5000; i > 0; i-=25) {
    memset(VFD_Buffer + 20, ' ', 5);             // Blanks where the number goes
    itoa(i,Num_Buff,10);
    VFD_Str_To_Buffer(VFD_Buffer+20+(5-strlen(Num_Buff)),Num_Buff);
    VFD_Send_Full_Buffer(VFD_Buffer, 0);
    delay(18);                                   // Delay less than decrement to keep on time!
  };
  
  //CPAGE--------------------------------------------------------------------------------------
  //  Blinking mode on corner characters.
  // - With the cursor off.
  //-------------------------------------------------------------------------------------------

  memset(VFD_Buffer,' ', sizeof(VFD_Buffer));
  VFD_Str_To_Buffer(VFD_Buffer,      "Blinking Stuff");
  VFD_Str_To_Buffer(VFD_Buffer+25,        "LOOK!");
  VFD_Send_Char(VFD_Mode_Cursor_Off);
  VFD_Send_Full_Buffer(VFD_Buffer, 0);

  // Insert some characters on the ends.
  VFD_Send_Cmd(20);                              // Cursor to R2C1
  VFD_Send_Char('X');
  VFD_Send_Cmd(39);                              // Cursor to R2C20
  VFD_Send_Char('X');

  delay(500);

  // Blink first and last columns of the 2nd line. 
  VFD_Send_Char(VFD_Mode_Blinking_On);           // Blink R2C1
  VFD_Send_Char(20);                             // - Posn to blink
  VFD_Send_Char(VFD_Mode_Blinking_On);           // Blink R2C20
  VFD_Send_Char(39);                             // - Posn to blink
  
  delay(2000);                                   // Gaze in wonder!

  // Turn off blinking mode. This cancels all blinking positions.
  VFD_Send_Char(VFD_Mode_Blinking_Off);

  delay(1000);

  //CPAGE--------------------------------------------------------------------------------------
  //  Move the cursor around a bunch with control characters and commands.
  //-------------------------------------------------------------------------------------------
  memset(VFD_Buffer, ' ', sizeof(VFD_Buffer)); 
  VFD_Str_To_Buffer(VFD_Buffer, "Cursor Moves:");
  VFD_Str_To_Buffer(VFD_Buffer + 20, "01234567890123456789");
  VFD_Send_Full_Buffer(VFD_Buffer,30);
  VFD_Send_Char(VFD_Mode_Cursor_On);
  delay(250);

  // Line Feed - move to the next line or back to the top.
  VFD_Str_To_Buffer(VFD_Buffer + 14, "LF");
  VFD_Send_Full_Buffer(VFD_Buffer,30);
  for (byte i = 0; i < 5; i++) {
    VFD_Send_Char(VFD_Ctl_LF);
    delay(250);
  };

  // Use the cursor move command to move the cursor around a bit just for fun..
  VFD_Str_To_Buffer(VFD_Buffer + 14, "MV");
  VFD_Send_Full_Buffer(VFD_Buffer,30);
  for (byte i=0; i < 20; i++) {
    VFD_Send_Cmd(bitRead(i,0)*20 + i);
    delay(250);
  };  
  delay(500);

  // Clear it and see that the cursor is not moved by a clear!
  VFD_Send_Cmd(10);
  VFD_Send_Char(VFD_Ctl_FF);
  VFD_Send_Str("Cursor not moved by FF!");
  delay(2000);
  
  //CPAGE--------------------------------------------------------------------------------------
  //  Horizontal scrolling.
  //-------------------------------------------------------------------------------------------
  const char LongMsg[] = "This is a demonstration of builtin scrolling on the VFD. Blanks are "
                         "sent at the end to clear the message.";

  memset(VFD_Buffer, ' ', sizeof(VFD_Buffer));
  VFD_Str_To_Buffer(VFD_Buffer, "Horizontal Scrolling");
  VFD_Send_Full_Buffer(VFD_Buffer, 0);
  VFD_Send_Cmd(20);                              
  VFD_Send_Char(VFD_Mode_HorzScroll);
  for (const char * pCh = LongMsg; *pCh != 0; *pCh++) {
    VFD_Send_Char(*pCh);
    delay(100);
  };

  // Scroll the message off with blanks.
  for (byte i = 0; i < 20; i++) {
    VFD_Send_Char(' ');
    delay(100);
  };
  VFD_Send_Char(VFD_Mode_Normal);
  
  //CPAGE--------------------------------------------------------------------------------------
  //  Vertical Scrolling.
  //-------------------------------------------------------------------------------------------
  memset(VFD_Buffer, ' ', sizeof(VFD_Buffer));
  VFD_Str_To_Buffer(VFD_Buffer, "Vertical Scrolling");
  VFD_Send_Full_Buffer(VFD_Buffer, 0);
  delay(1000);
  VFD_Send_Char(VFD_Mode_VertScroll);
  for (byte i = 0; LongMsg[i] != 0; i++) {
    VFD_Send_Char(LongMsg[i]);
    delay(50);
  };

  // Scroll the message off with blanks.
  for (byte i = 0; i < 40; i++) {
    VFD_Send_Char(' ');
    delay(50);
  };
  VFD_Send_Char(VFD_Mode_Normal);

  //CPAGE--------------------------------------------------------------------------------------
  //  Software Reset
  //-------------------------------------------------------------------------------------------
  VFD_Send_Cmd(VFD_Cmd_SWReset);                 // Software reset
  delay(1000);                                   // S/B a blank display
  
  //CPAGE--------------------------------------------------------------------------------------
  //  Show all of the printable characters and their codes.
  //-------------------------------------------------------------------------------------------
  byte Len;

  VFD_Send_Char(0x0C);                             // Clear the display
  VFD_Send_Cmd(0);
  for (int Ascii=32; Ascii <= 255; Ascii++) {     
    Len=sprintf(VFD_Buffer,"Ascii: %02X %3d",Ascii,Ascii);
    
    // Put our message on line 1
    VFD_Send_Cmd(0);        // Force it to position 0    
    for (byte i=0;  i < Len; i++) VFD_Send_Char(VFD_Buffer[i]); 
    for (byte i=Len; i < 20; i++) VFD_Send_Char(' ');
    
    // And on line 2 - This will be erased if the character actually displays anything.
    VFD_Send_Cmd(20);       // Force it to position 20
    for (byte i = 0; i < 20;  i++) VFD_Send_Char('A' + i);
    
    // Now write a line of the Ascii character to the second line
    // - Non-displaying characters do not display anything and do not move the cursor.
    VFD_Send_Cmd(20);       // Force it to position 20
    for (byte i=0; i < 20; i++) {
      VFD_Send_Char(Ascii);      // Send some good stuff to the VFD!
    };
    delay(250);  
  };

  //CPAGE--------------------------------------------------------------------------------------
  //  Display various bitmaps in a custom character.
  //-------------------------------------------------------------------------------------------

  // Try out some custom characters. We can only have two custom bitmaps at a time.
  // - We will change the bit map for 'A'
  // There is a bug here. The VFD uses two buffers for character bitmaps.
  // A write will always use the next buffer, even if it is re-defining a character that already
  // has a bit map assigned. The implications are: 
  // - Redefining a single character will erase the definition for another character.
  // - Here's the bug: If you redefine an overridden character, it does not update the
  // - Character to use the new buffer so you don't see the update. If you define two characters
  //   or re-define the same character twice it works ok.
  // Bottom line is this does not work well for dynamically changing characters. If you load two
  // custom character bitmaps during initialization and then use them, it will be fine.
  byte Data[5];

  memset(VFD_Buffer, ' ', sizeof(VFD_Buffer));
  VFD_Str_To_Buffer(VFD_Buffer, "Custom bitmaps.");
  VFD_Send_Full_Buffer(VFD_Buffer, 0);
  VFD_Send_Char(VFD_Mode_Cursor_Off);            // No cursor!

  // Create custom bitmaps with one additional bit turned on each time.
  // - From this display, you can see how the bits are translated to pixels. 
  memset(Data,0x00,sizeof(Data));                // Clear all of the data
  for (byte row = 0; row < 5; row++) {
    for (byte bt=0; bt < 8; bt++) {
      bitSet(Data[row],bt);                      // Set another bit on
      VFD_Send_Char(VFD_Set_Char_Bitmap);        // Send ESC
      VFD_Send_Char('A');                        // Send character to modify
      VFD_Send_Char(Data[0]);
      VFD_Send_Char(Data[1]);
      VFD_Send_Char(Data[2]);
      VFD_Send_Char(Data[3]);
      VFD_Send_Char(Data[4]);

      // twice because of double-buffer bug. The buffers will now both be the same.
      VFD_Send_Char(VFD_Set_Char_Bitmap);        // Send ESC
      VFD_Send_Char('A');                        // Send character to modify
      VFD_Send_Char(Data[0]);
      VFD_Send_Char(Data[1]);
      VFD_Send_Char(Data[2]);
      VFD_Send_Char(Data[3]);
      VFD_Send_Char(Data[4]);

      VFD_Send_Cmd(20);                          // Move to posn 20
      VFD_Send_Char('B');   
      VFD_Send_Char('0'+row);   
      VFD_Send_Char('b');   
      VFD_Send_Char('0'+bt);   

      // Display a few of the new characters.
      for (byte i=0; i < 10; i++) {
        VFD_Send_Char('A');   
      };
      delay(250);                                // Slow it all down so we can see it.
    };
  };
  
  delay(1000);                                   // Gaze in wonder again!
  
  //CPAGE--------------------------------------------------------------------------------------
  //  Since we've got a custom bitmap will all bits on, lets light them all up at once!
  //-------------------------------------------------------------------------------------------
  memset(VFD_Buffer,'A',40);
  VFD_Send_Full_Buffer(VFD_Buffer,0);
  delay(2000);
  
  // Another SW reset test. All the character should be normal again after this.
  VFD_Send_Cmd(VFD_Cmd_SWReset);                 // Software reset
  VFD_Send_Char(VFD_Mode_Cursor_Off);            // No cursor!

  //CPAGE--------------------------------------------------------------------------------------
  //  Change the display brightness.
  //-------------------------------------------------------------------------------------------
  memset(VFD_Buffer, ' ', sizeof(VFD_Buffer));
  VFD_Str_To_Buffer(VFD_Buffer, "Brightness: X");
  VFD_Str_To_Buffer(VFD_Buffer + 20, "ABCDEFGHIJKLMNOPQRST");
  VFD_Send_Full_Buffer(VFD_Buffer, 0);

  // Change the intensity to teh four supported value.
  // -  0, 16, 32, 48.
  // -  0 is brightest.
  // - 48 is dimmest.
  for (byte num=0; num < 4; num++) {  // Run through it a few times!
    for (byte i=0; i <= 3; i++) {
      VFD_Send_Char(VFD_Set_Brightness);         // intensity
      VFD_Send_Char(i<<4);                       // Send code (mult by 16)
      VFD_Send_Cmd(12);                          // Put 0-3 in the header 
      VFD_Send_Char('0' + i);
      delay(1000);
    };  
  };
  VFD_Send_Char(VFD_Set_Brightness);             // Back to full intensity.
  VFD_Send_Char(0);     
  
}; // Loop

//CPAGE========================================================================================
// VFD_Str_To_Buffer
//
// - Copy the string in the second parameter into Buff.
// - Note that this is not a string copy in that the termnating NULL is NOT copied, but is 
//   required to be in the string.
// 
//=============================================================================================
void VFD_Str_To_Buffer(char * Buff, char * Str) {

while (*Str != 0) {
  *Buff=*Str;
  *Buff++;
  *Str++;
};

};  //  VFD_Str_To_Buffer

//CPAGE========================================================================================
// VFD_Send_Full_Buffer(char * Buff, byte CurPosn);
//
// - Here is the function to use for Displays-for-Dummies.
// - Format all the characters you want to see in a 40-byte buffer.
//   - The buffer should only contain 'printable' characters for the disaplay.
//   - No control characters, or other characters that do not display.
// - Hand it over to this function and the display will be updated appropriately.
// - The cursor mode is not changed. That is, if it was displayed, it will remain displayed.
// - The cursor is left in the position indicated by the second parameter.
// - If blinking characters are set, the blinking is not changed.
// - You should not call this function if vertical or horizontal scrolling mode is turned on:
//   - It is not possible to write a character in the last column without forcing a
//     horizontal scroll.
//   - A vertical scroll will happen if the last column on line two is written to.
// - If you need scrolling on, this routine is not for you!
// - This might seem like over kill if you are only changing a couple of characters on
//   the display, but the full refresh takes less than 5 milliseconds and makes the 
//   display management easy.
// 
//=============================================================================================
void VFD_Send_Full_Buffer(char Buff[], byte CurPosn) {

  VFD_Send_Cmd(0);                               // Move cursor to beginning 
  for (byte i = 0; i < 40; i++) {
    VFD_Send_Char(Buff[i]);                      // Write out the full buffer
  };
  if (CurPosn > 39) {
    CurPosn = 0;                                 // Fix possible bad value.
  };
  VFD_Send_Cmd(CurPosn);                         // Put the cursor in position.
  
};  //  VFD_Send_Full_Buffer

//CPAGE========================================================================================
// VFD_Send_Str
//
// - Send all of the character in a string to the VFD.
// - They are written starting at the current cursor position.
// 
//=============================================================================================
void VFD_Send_Str(char * Str) {

while (*Str != 0) {
  VFD_Send_Char(*Str);
  *Str++;
};

};  //  VFD_Send_Str

//CPAGE========================================================================================
// VFD_Send_Cmd and VFD_Send_Char
//
// - Commands are sent with AD high. Data are sent with AD low.
// - Send a command or data character to the display.
// - Direct port manipulation would be a good idea here if speed matters at all. It does not
//   for this sample application. 
// - This code supports any line to be on just about any digital capable pin of the Arduino. 
// - The digitalWrite's and pinMode's below each take about 4us to execute for a total of
//   about 45-50us total to run this function. 
// - It could probably be done in 1-2us if D0-D7 were on a single port,and all control lines
//   on anotehr port.
// - The display is happy with sub-microsecond switching times on its pins. 
//
// Specific to VFD_Send_Cmd:
//
// - The AD line must be implemented for Send_Cmd to work.
// - A parameters of 0-39 will execute a move-cursor command to place the cursor on the given
//   positon of the display. Over 39 will go to position 0.
// - A parameter of 0x40 (64) will perform a software initiated reset of the display.
// 
//=============================================================================================
void VFD_Send_Cmd(byte Ch) {
  
  VFD_Check_Cmd_Timing();                    // Can we issue the next command yet?

  // Issue command to VFD
  digitalWrite(VFD__WR, LOW);                   //Prepare to write
  digitalWrite(VFD_AD, HIGH);                   // Set command mode
  digitalWrite(VFD_D0, bitRead(Ch,0));
  digitalWrite(VFD_D1, bitRead(Ch,1));
  digitalWrite(VFD_D2, bitRead(Ch,2));
  digitalWrite(VFD_D3, bitRead(Ch,3));
  digitalWrite(VFD_D4, bitRead(Ch,4));
  digitalWrite(VFD_D5, bitRead(Ch,5));
  digitalWrite(VFD_D6, bitRead(Ch,6));
  digitalWrite(VFD_D7, bitRead(Ch,7));
  digitalWrite(VFD__WR, HIGH); // Write Complete

  LastCmdTime=micros();                          // Rememeber when the last command was issued.

  // Set the delay until the next command can be issued.
  if (Ch == VFD_Cmd_SWReset) {                   
    VFDCmdWait = VFD_Delay_SWReset;
  }  else {                                      // Assume a cursor move.
    VFDCmdWait = VFD_Delay_Cursor_Posn;
  };

};  //  VFD_Send_Cmd

//CPAGE========================================================================================
// VFD_Send_Char
//
// - See VFD_Send_Cmd for more inoformation.
//
// Specific to VFD_Send_Char:
//
// - The follwoing parameter values are supported. See the device documentation for details.
//   - 0x00 - 0x1F - Various control commands that set operation mode and can define characters.
//                 - Some commands are multi-character so they will continue to process the 
//                   data that follows as parameters to the command. 
//   - 0x20 - 0xff - Displays the character from the character generation rom at the current 
//                   cursor position. Not all character positions are defined in the ROM.
//=============================================================================================
void VFD_Send_Char(byte Ch) {

  VFD_Check_Cmd_Timing();                        // Can we issue the next command yet?

  // Issue command to VFD
  digitalWrite(VFD__WR, LOW);                    //Prepare to write
  digitalWrite(VFD_AD, LOW);                     // Set data mode
  digitalWrite(VFD_D0, bitRead(Ch,0));           // Put the bits on the data lines
  digitalWrite(VFD_D1, bitRead(Ch,1));
  digitalWrite(VFD_D2, bitRead(Ch,2));
  digitalWrite(VFD_D3, bitRead(Ch,3));
  digitalWrite(VFD_D4, bitRead(Ch,4));
  digitalWrite(VFD_D5, bitRead(Ch,5));
  digitalWrite(VFD_D6, bitRead(Ch,6));
  digitalWrite(VFD_D7, bitRead(Ch,7));
  digitalWrite(VFD__WR, HIGH);                   // Write Complete

  LastCmdTime=micros();                          // Remember when the last command was issued.

  // Set the delay until the next command can be issued.
  if (Ch == VFD_Ctl_FF) {
    VFDCmdWait = VFD_Delay_Clear;               // Clears take a lot longer!
  } else {                                      
    VFDCmdWait = VFD_Delay_Mode_Normal;         // Assume a worst case command
  };
  
};  //  VFD_Send_Char

//CPAGE========================================================================================
// VFD_HW_Reset
//
// - Perform a hardware reset command on the VFD.
// - Just pulse the reset line high for a few us.
//   - About 4.0us was enough (that's about as fast as the Arduino can toggle the line using 
//     digitalWrite). But bringing it up to a total of about 10us will make sure the pulse
//     did not get sucked up by some capacatence or inductance somewhere.
// After reset:
// - The display is cleared
// - The cursor position is set to 0 (top left
// - Cursor mode DC1 is set (visible, non-flashing cursor)
// - Scroll modes are off.
// TODO:
// - How long does a reset really take? Need to monitor the Status line.
//=============================================================================================
void VFD_HW_Reset() {

  VFD_Check_Cmd_Timing();                        // Can we issue the next command yet?

  digitalWrite(VFD_RS, HIGH);                    // Pulse the reset line
  delayMicroseconds(VFD_Delay_Reset_Pulse);                 
  digitalWrite(VFD_RS, LOW);                     // Done with the pulse

  LastCmdTime = micros();                        // Remember when the last command was issued.
  VFDCmdWait = VFD_Delay_HWReset;                // Don't issue the next command for this long.

};  //  VFD_HW_Reset

//CPAGE========================================================================================
// VFD_Test_Mode
//
// - Put the VFD in test mode.
//   - Steps the display through all of the characters it supports.
//   - It will stay in this mode until it is powered off or HW reset. 
//   - Control sequence:
//     - _TS=0
//     - Short positive reset pulse. (4.0us was OK)
//     - Hold _TS=0 for at least 130ms. 129ms did not work. Use 150ms to be sure.
//     - _TS=1
// - The display will stay in test mode until the next hardware reset or power cycle.
//   - It displays its internal character set from 0x20-0xFF on the display at a rate of about
//     150ms/char. When its done, it just does it again.
//   - One run through the sequence takes about 28 seconds.
//=============================================================================================
void VFD_Test_Mode() {

  VFD_Check_Cmd_Timing();                        // Can we issue the next command yet?
  
  digitalWrite(VFD__TS,LOW);                     // Pull down the test line 
  delayMicroseconds(VFD_Delay_Reset_Pulse); 
  digitalWrite(VFD_RS,HIGH);                     // Call for a reset
  delayMicroseconds(VFD_Delay_Reset_Pulse); 
  digitalWrite(VFD_RS,LOW);                      // Reset pulse is done
  delay(VFD_Hold_Test/1000);                     // Hold down _TS until the test starts
  digitalWrite(VFD__TS, HIGH);                   // Can release test now

  LastCmdTime = micros();                        // Rememeber when the last command was issued.
  VFDCmdWait = VFD_Delay_Test;                   // Don't issue the next command for this long.

};  //  VFD_Test_Mode

//CPAGE========================================================================================
// VFD_Check_Cmd_Timing
//
// - Since the VFD does not return any sort of "busy" signal, we must wait a certain amount
//   of time before issuing another command to it. 
// - If not enough time has passed, we will delay for the remainder of the command-busy time.
// - The time we wait is dependent on the actual command that was issued. That time is 
//   set in a global whenever a command is issued.
// - The micros() function rolls over about every 70 minutes (16Mhz Arduino):
//   - It is possible that over 70 minutes has passed since a command was last issued.
//   - In a tiny number of cases we will end up inserting a few more microseconds of wait time, 
//     but it won't matter as we've already been waiting a very long time. 
//=============================================================================================
void VFD_Check_Cmd_Timing() {

  unsigned long     DeltCmdTime = 0;             // us since the last commnd was executed.

  DeltCmdTime = micros() - LastCmdTime;          // How long since we issued the last command?
  if (DeltCmdTime < VFDCmdWait) {                // Do we still need to wait for a while?
    DeltCmdTime=VFDCmdWait - DeltCmdTime;        // How long to wait 
    if (DeltCmdTime > 15000) {                   // delaymicroseconds does not work above 16000!
      delay(DeltCmdTime/1000);                   // wait in milliseonds
    } else {  
      delayMicroseconds(DeltCmdTime); // Wait the required microseconds
    };
   };

};  // VFD_Check_Cmd_Timing  

